#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 21 21:13:23 2019

@author: tommy
"""

import operator


def func_addition(a, b):
    """Permforms addition."""
    return operator.add(a, b)


def func_subtraction(a, b):
    """Permforms addition."""
    return operator.subtract(a, b)


class BinaryOperator:
    def __init__(self, operator):
        self.operator = operator

    def get_operator(self):
        return self.operator

    def set_operator(self, operator):
        self.operator = operator



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module is the entrypoint when a user calls:
$ python -m treedoc [object]
"""

from treedoc.main import CLI_entrypoint

if __name__ == "__main__":
    CLI_entrypoint()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 21 21:13:23 2019

@author: tommy
"""

import functools
import operator

from treedoctestpackage.module2 import SuperClass
from treedoctestpackage.module2 import function_nested_outer as imported_function
from treedoctestpackage.module2 import wrapper


class SubClass(SuperClass):
    """Test that the SuperClass will not be found in this module, but in it's original."""

    def subclass_method(self):
        return 1


def func_addition(a, b):
    """Permforms addition."""
    return operator.add(a, b)


def func_using_imported(x):
    return imported_function(x)


add_five_partial = functools.partial(func_addition, a=5)


def func_many_args(a, b=2, c=4, d=(1, 2, 3)):
    """Function with many arguments."""
    return b + c


def func_many_long_args(
    arg=2,
    num: int = 123,
    name: str = "john",
    pi: float = 3.14,
    place: str = "london",
    e: float = 2.718281828459045,
):
    """Function with many arguments."""
    return None


def generator(a):
    for i in range(a):
        yield i


# =============================================================================
# WRAPPED USING A WRAPPER FROM AN OUTSIDE MODULE
# =============================================================================


@wrapper
def func_wrapped_w_wrapper_from_other_module(x):
    return x


# =============================================================================
# Classes
# =============================================================================


class MyClass:
    def __init__(self):
        pass

    def method_bound_to_myclass(self, a, b: int, *args, c=4.2, d: int = 42, **kwargs):
        """Method docstring."""
        return a

    @classmethod
    def classmethod_bound_to_myclass(
        cls, a, b: int, *args, c=4.2, d: int = 42, **kwargs
    ):
        """Class method docstring."""
        return a

    @staticmethod
    def static_method_bound_to_myclass(
        self, a, b: int, *args, c=4.2, d: int = 42, **kwargs
    ):
        """Static method docstring."""
        return a

    def __add__(self, other):
        return type(self)()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 21 21:14:43 2019

@author: tommy
"""

from treedoctestpackage.module import MyClass
from treedoctestpackage.subpackage.subpackagemodule import func_subtraction

MyClass = MyClass
func_subtraction = func_subtraction



from setuptools import find_packages, setup

setup(
    name="treedoctestpackage",
    version="0.0.1",
    description="A test package for treedoc.",
    long_description="A test package for treedoc.",
    author="tommyod",
    author_email="tod001@uib.no",
    license="MIT",
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=[],
    # https://python-packaging.readthedocs.io/en/latest/command-line-scripts.html
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
    ],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test the command line interface (CLI).
"""
import itertools
import random
import subprocess

import pytest

object_strings = [
    "list",
    "collections",
    "collections.Counter",
    "math",
    "collections.abc.Collection",
    "list set dict",
    "list   set dict  ",
]


@pytest.mark.parametrize("obj_string", object_strings)
def test_cli_smoketests(obj_string):
    """The smoketests assure that the commands run without errors. No output testing
    is performed apart from the non-existence of an error."""

    exit_code, output = subprocess.getstatusoutput(" ".join(["treedoc", obj_string]))

    # Zero exit code means everything is OK
    assert exit_code == 0


packages = ["numpy", "scipy", "pandas", "sklearn"]


@pytest.mark.parametrize("library", packages)
def test_cli_smoketests_famous_packages(library):
    """The smoketests assure that the commands run without errors. No output testing
    is performed apart from the non-existence of an error."""

    args = " ".join(["treedoc", library, "--subpackages"])
    exit_code, output = subprocess.getstatusoutput(args)

    # Zero exit code means everything is OK
    assert exit_code == 0


def _generate_cli_args(n):
    """Generate n args."""
    random.seed(42)

    yielded = set()

    for _ in range(n):

        to_yield = " ".join(
            [
                "--level " + str(random.choice([1, 2, 3])),
                random.choice(["", "--subpackages"]),
                random.choice(["", "--modules"]),
                random.choice(["", "--private"]),
                random.choice(["", "--tests"]),
                random.choice(["--printer tree", "--printer dense"]),
                "--signature " + str(random.choice([0, 1, 2, 3, 4])),
                "--docstring " + str(random.choice([0, 1, 2])),
                "--info " + str(random.choice([0, 1, 2])),
            ]
        )
        if to_yield not in yielded:
            yielded.add(to_yield)
            yield to_yield


@pytest.mark.parametrize(
    "arg_string",
    [
        " ".join([i, j])
        for i, j in itertools.product(object_strings, _generate_cli_args(2))
    ],
)
def test_cli_smoketests_w_args(arg_string):
    """The smoketests assure that the commands run without errors. No output testing
    is performed apart from the non-existence of an error."""

    exit_code, output = subprocess.getstatusoutput(" ".join(["treedoc", arg_string]))
    assert exit_code == 0


if __name__ == "__main__":
    pytest.main(args=[__file__, "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
General utilities for treedoc.
"""

import collections
import os
import typing


class PrintMixin:
    """Adds a __repr__ method to a class."""

    # Could've used dataclasses, but they were introduced in Python 3.8

    def __repr__(self) -> str:
        """Returns a printable representation of an object, e.g. 'ClassName(a=2, b=3)'."""
        args = ["{}={}".format(k, v) for k, v in self.__dict__.items()]
        return type(self).__name__ + "({})".format(", ".join(args))


_marker = object()


class Peekable:
    """Wrap an iterator to allow lookahead.
    
    Call `peek` on the result to get the value that will be returned by `next`. 
    This won't advance the iterator:
        
    >>> p = Peekable(['a', 'b'])
    >>> p.peek()
    'a'
    >>> next(p)
    'a'
        
    Pass `peek` a default value to return instead of raising ``StopIteration`` 
    when the iterator is exhausted.
    
    >>> p = Peekable([])
    >>> p.peek('hi')
    'hi'
    """

    def __init__(self, iterable):
        self._it = iter(iterable)
        self._cache = collections.deque()

    def __iter__(self):
        return self

    def __bool__(self) -> bool:
        try:
            self.peek()
        except StopIteration:
            return False
        return True

    def peek(self, default=_marker):
        """Return item that will be returned from ``next()``.
        
        Returns ``default`` if iterator is exhausted. If ``default`` is not
        provided, raise ``StopIteration``.
        """
        if not self._cache:
            try:
                self._cache.append(next(self._it))
            except StopIteration:
                if default is _marker:
                    raise
                return default
        return self._cache[0]

    def __next__(self):
        if self._cache:
            return self._cache.popleft()

        return next(self._it)


def get_terminal_size(fallback=(128, 24)) -> typing.Tuple[int, int]:
    """Get the terminal size.
    
    See http://granitosaurus.rocks/getting-terminal-size.html
    """
    for i in range(0, 3):
        try:
            columns, rows = os.get_terminal_size(i)
        except OSError:
            continue
        break
    else:  # set default if the loop completes which means all failed
        columns, rows = fallback
    return columns, rows



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Tests for classes and functions located in `printing.py`.
"""

import builtins
import collections.abc
import datetime
import itertools
import math
import operator
from collections.abc import Callable

import pytest

import treedoctestpackage
from treedoc.printing import (
    TreePrinter,
    format_signature,
    get_docstring,
    resolve_input,
    resolve_str_to_obj,
    signature_from_docstring,
)
from treedoc.utils import Peekable


class TestTreePrinterRowFormatting:
    def test_row_formatting_ex1(self):
        """Test on an example drawn by hand."""

        printer = TreePrinter()

        # These can be substituted for further generalization.
        # TODO
        # =============================================================================
        #         RIGHT = "├──"
        #         DOWN = "│  "
        #         LAST = "└──"
        #         BLANK = "   "
        # =============================================================================

        rows = [
            (["root"], [True]),
            (["root", "A"], [True, False]),
            (["root", "A", "a"], [True, False, False]),
            (["root", "A", "b"], [True, False, False]),
            (["root", "A", "c"], [True, False, True]),
            (["root", "B"], [True, True]),
            (["root", "B", "a"], [True, True, False]),
            (["root", "B", "b"], [True, True, True]),
        ]

        rows = Peekable(iter(rows))
        formatted_rows = list(printer._format_row(rows))

        assert formatted_rows == [
            ([""], ["root"]),
            (["", "├──"], ["root", "A"]),
            (["", "│  ", "├──"], ["root", "A", "a"]),
            (["", "│  ", "├──"], ["root", "A", "b"]),
            (["", "│  ", "└──"], ["root", "A", "c"]),
            (["", "└──"], ["root", "B"]),
            (["", "   ", "├──"], ["root", "B", "a"]),
            (["", "   ", "└──"], ["root", "B", "b"]),
        ]

    def test_row_formatting_ex2(self):
        """Test on an example drawn by hand."""

        printer = TreePrinter()

        rows = [
            (["root"], [True]),
            (["root", "A"], [True, False]),
            (["root", "A", "a"], [True, False, False]),
            (["root", "A", "b"], [True, False, True]),
            (["root", "A", "b", "A"], [True, False, True, False]),
            (["root", "A", "b", "B"], [True, False, True, False]),
            (["root", "A", "b", "C"], [True, False, True, True]),
            (["root", "B"], [True, True]),
            (["root", "B", "a"], [True, True, False]),
            (["root", "B", "b"], [True, True, True]),
        ]

        rows = Peekable(iter(rows))
        formatted_rows = list(printer._format_row(rows))

        assert formatted_rows == [
            ([""], ["root"]),
            (["", "├──"], ["root", "A"]),
            (["", "│  ", "├──"], ["root", "A", "a"]),
            (["", "│  ", "└──"], ["root", "A", "b"]),
            (["", "│  ", "   ", "├──"], ["root", "A", "b", "A"]),
            (["", "│  ", "   ", "├──"], ["root", "A", "b", "B"]),
            (["", "│  ", "   ", "└──"], ["root", "A", "b", "C"]),
            (["", "└──"], ["root", "B"]),
            (["", "   ", "├──"], ["root", "B", "a"]),
            (["", "   ", "└──"], ["root", "B", "b"]),
        ]

    def test_row_formatting_ex3(self):
        """Test on an example drawn by hand."""

        printer = TreePrinter()

        rows = [
            (["root"], [True]),
            (["root", "A"], [True, False]),
            (["root", "B"], [True, False]),
            (["root", "C"], [True, False]),
            (["root", "C", "a"], [True, False, False]),
            (["root", "C", "b"], [True, False, False]),
            (["root", "C", "c"], [True, False, False]),
            (["root", "C", "c", "A"], [True, False, False, False]),
            (["root", "C", "c", "B"], [True, False, False, False]),
            (["root", "C", "c", "C"], [True, False, False, False]),
            (["root", "C", "c", "D"], [True, False, False, False]),
            (["root", "C", "c", "D", "a"], [True, False, False, False, False]),
            (["root", "C", "c", "D", "b"], [True, False, False, False, False]),
            (["root", "C", "c", "D", "c"], [True, False, False, False, False]),
            (["root", "C", "c", "D", "d"], [True, False, False, False, False]),
            (
                ["root", "C", "c", "D", "d", "A"],
                [True, False, False, False, False, False],
            ),
            (
                ["root", "C", "c", "D", "d", "B"],
                [True, False, False, False, False, False],
            ),
            (
                ["root", "C", "c", "D", "d", "C"],
                [True, False, False, False, False, True],
            ),
            (
                ["root", "C", "c", "D", "d", "C", "a"],
                [True, False, False, False, False, True, True],
            ),
            (["root", "C", "c", "D", "e"], [True, False, False, False, True]),
            (
                ["root", "C", "c", "D", "e", "A"],
                [True, False, False, False, True, True],
            ),
            (["root", "C", "c", "E"], [True, False, False, True]),
            (["root", "C", "c", "E", "a"], [True, False, False, True, False]),
            (["root", "C", "c", "E", "b"], [True, False, False, True, False]),
            (["root", "C", "c", "E", "c"], [True, False, False, True, False]),
            (["root", "C", "c", "E", "d"], [True, False, False, True, True]),
            (["root", "C", "d"], [True, False, False]),
            (["root", "C", "e"], [True, False, True]),
            (["root", "D"], [True, True]),
            (["root", "D", "a"], [True, True, False]),
            (["root", "D", "b"], [True, True, False]),
            (["root", "D", "c"], [True, True, False]),
            (["root", "D", "d"], [True, True, False]),
            (["root", "D", "e"], [True, True, True]),
        ]

        rows = Peekable(iter(rows))
        formatted_rows = list(printer._format_row(rows))

        assert formatted_rows == [
            ([""], ["root"]),
            (["", "├──"], ["root", "A"]),
            (["", "├──"], ["root", "B"]),
            (["", "├──"], ["root", "C"]),
            (["", "│  ", "├──"], ["root", "C", "a"]),
            (["", "│  ", "├──"], ["root", "C", "b"]),
            (["", "│  ", "├──"], ["root", "C", "c"]),
            (["", "│  ", "│  ", "├──"], ["root", "C", "c", "A"]),
            (["", "│  ", "│  ", "├──"], ["root", "C", "c", "B"]),
            (["", "│  ", "│  ", "├──"], ["root", "C", "c", "C"]),
            (["", "│  ", "│  ", "├──"], ["root", "C", "c", "D"]),
            (["", "│  ", "│  ", "│  ", "├──"], ["root", "C", "c", "D", "a"]),
            (["", "│  ", "│  ", "│  ", "├──"], ["root", "C", "c", "D", "b"]),
            (["", "│  ", "│  ", "│  ", "├──"], ["root", "C", "c", "D", "c"]),
            (["", "│  ", "│  ", "│  ", "├──"], ["root", "C", "c", "D", "d"]),
            (
                ["", "│  ", "│  ", "│  ", "│  ", "├──"],
                ["root", "C", "c", "D", "d", "A"],
            ),
            (
                ["", "│  ", "│  ", "│  ", "│  ", "├──"],
                ["root", "C", "c", "D", "d", "B"],
            ),
            (
                ["", "│  ", "│  ", "│  ", "│  ", "└──"],
                ["root", "C", "c", "D", "d", "C"],
            ),
            (
                ["", "│  ", "│  ", "│  ", "│  ", "   ", "└──"],
                ["root", "C", "c", "D", "d", "C", "a"],
            ),
            (["", "│  ", "│  ", "│  ", "└──"], ["root", "C", "c", "D", "e"]),
            (
                ["", "│  ", "│  ", "│  ", "   ", "└──"],
                ["root", "C", "c", "D", "e", "A"],
            ),
            (["", "│  ", "│  ", "└──"], ["root", "C", "c", "E"]),
            (["", "│  ", "│  ", "   ", "├──"], ["root", "C", "c", "E", "a"]),
            (["", "│  ", "│  ", "   ", "├──"], ["root", "C", "c", "E", "b"]),
            (["", "│  ", "│  ", "   ", "├──"], ["root", "C", "c", "E", "c"]),
            (["", "│  ", "│  ", "   ", "└──"], ["root", "C", "c", "E", "d"]),
            (["", "│  ", "├──"], ["root", "C", "d"]),
            (["", "│  ", "└──"], ["root", "C", "e"]),
            (["", "└──"], ["root", "D"]),
            (["", "   ", "├──"], ["root", "D", "a"]),
            (["", "   ", "├──"], ["root", "D", "b"]),
            (["", "   ", "├──"], ["root", "D", "c"]),
            (["", "   ", "├──"], ["root", "D", "d"]),
            (["", "   ", "└──"], ["root", "D", "e"]),
        ]


def map_itemgetter(iterable, index: int):
    """Map an itemgetter over an iterable, returning element correpoding to index."""
    getter = operator.itemgetter(index)
    for item in iterable:
        yield getter(item)


@pytest.mark.parametrize(
    "input_arg, expected",
    [
        (math.log, ("x[, base]", "x, [base=math.e]")),
        (dict.pop, ("k[,d]", None)),
        (datetime.datetime.strptime, (None,)),
    ],
)
def test_signature_from_docstring(input_arg, expected):
    """Test the function getting signature information from docstrings.
    
    Due to builtin docstrings differing on different Python versions,
    we allow several possibilities."""
    assert signature_from_docstring(input_arg) in expected


def test_get_docstring():
    """Test retrieval of docstrings."""

    def func():
        pass

    return_val = "This is the docstring."

    func.__doc__ = """This is the docstring."""
    assert get_docstring(func) == return_val

    func.__doc__ = """
    
    
    This is the docstring.
    
    
    """
    assert get_docstring(func) == return_val

    func.__doc__ = """
    This is the docstring.
    
    This is more stuff.
    """
    assert get_docstring(func) == return_val

    func.__doc__ = """
    This is the docstring. More information here.
    
    Even more stuff.
    """
    assert get_docstring(func) == "This is the docstring. More information here."
    assert get_docstring(func, width=12) == "This is..."

    delattr(func, "__doc__")
    assert get_docstring(func) == ""

    func.__doc__ = """
    Lorem Ipsum is simply dummy text of the printing and typesetting industry. 
    Lorem Ipsum has been the industry's standard dummy text ever since the 1500s, 
    when an unknown printer took a galley of type and scrambled it to make a type 
    specimen book. It has survived not only five centuries, but also the leap into 
    electronic typesetting, remaining essentially unchanged. It was popularised 
    in the 1960s with the release of Letraset sheets containing Lorem Ipsum passages, 
    and more recently with desktop publishing software like Aldus PageMaker 
    including versions of Lorem Ipsum.
    """

    assert get_docstring(func, width=16) == "Lorem Ipsum..."


class TestObjectResolution:
    @staticmethod
    @pytest.mark.parametrize(
        "input_arg, expected",
        [
            ("list", list),
            ("builtins.set", set),
            ("operator.add", operator.add),
            ("collections.abc.Callable", Callable),
            ("collections.abc", collections.abc),
            ("", builtins),
        ],
    )
    def test_resolve_str_to_obj(input_arg, expected):
        """
        Test that objects are resolved from strings as expected.
        """
        assert resolve_str_to_obj(input_arg) == expected

    @staticmethod
    def test_resolve_str_to_obj_raises():

        with pytest.raises(ImportError):
            resolve_str_to_obj("gibberish")

    @staticmethod
    @pytest.mark.parametrize(
        "input_arg, expected",
        [
            ("list dict set", [list, dict, set]),
            (["list", "dict", "set"], [list, dict, set]),
            (["list", "dict", set], [list, dict, set]),
            ([list, dict, set], [list, dict, set]),
            ("list dict", [list, dict]),
            ("list   dict ", [list, dict]),
            ("list", [list]),
        ],
    )
    def test_resolve_input(input_arg, expected):
        """
        Test that objects are resolved from inputs.
        """

        assert resolve_input(input_arg) == expected

    @staticmethod
    def test_resolve_input_raises():

        with pytest.raises(ImportError):
            resolve_input("list dict gibberish")


class TestSignature:
    """
    Class for gathering format_signature() tests. 
    Note the stripping of whitespaces in some of the tests. More info on this in PR #9.
    """

    parameters = [
        (0, ""),
        (1, "(...)"),
        (2, "(a,b,*args,c,d,**kwargs)"),
        (3, "(a,b,*args,c=4.2,d=42,**kwargs)"),
        (4, "(a,b:int,*args,c=4.2,d:int=42,**kwargs)"),
    ]

    @staticmethod
    @pytest.mark.parametrize("verbosity, expected", parameters)
    def test_keywords_annotated_defaults_args_kwargs(verbosity, expected):
        """ 
        Test that formatting signature works on a user defined function.
        """

        def myfunc1(a, b: int, *args, c=4.2, d: int = 42, **kwargs):
            return None

        assert (
            "".join(
                char
                for char in format_signature(myfunc1, verbosity=verbosity)
                if char != " "
            )
            == expected
        )

    @staticmethod
    @pytest.mark.parametrize(
        "verbosity, expected", [(0, ""), (1, "()"), (2, "()"), (3, "()"), (4, "()")]
    )
    def test_empty_signature(verbosity, expected):
        """ 
        Test that formatting signature works on a user defined function with no arguments.
        """

        def myfunc2():
            return None

        assert format_signature(myfunc2, verbosity=verbosity) == expected

    @staticmethod
    @pytest.mark.parametrize(
        "verbosity, expected",
        [
            (0, ""),
            (1, "(...)"),
            (2, "(self, n)"),
            (3, "(self, n=None)"),
            (4, "(self, n=None)"),
        ],
    )
    def test_builtin_class(verbosity, expected):
        """ 
        Test that formatting signature works on a built-in class.
        """
        from collections import Counter

        assert format_signature(Counter.most_common, verbosity=verbosity) == expected

    @staticmethod
    @pytest.mark.parametrize("verbosity, expected", parameters)
    def test_method(verbosity, expected):
        """
        Test that formatting signature works on a method.
        """
        myclass = treedoctestpackage.MyClass()

        assert (
            "".join(
                char
                for char in format_signature(
                    myclass.method_bound_to_myclass, verbosity=verbosity
                )
                if char != " "
            )
            == expected
        )

    @staticmethod
    @pytest.mark.parametrize(
        "verbosity, expected",
        [
            (0, ""),
            (1, "(...)"),
            (2, "(self,a,b,*args,c,d,**kwargs)"),
            (3, "(self,a,b,*args,c=4.2,d=42,**kwargs)"),
            (4, "(self,a,b:int,*args,c=4.2,d:int=42,**kwargs)"),
        ],
    )
    def test_static_method(verbosity, expected):
        """
        Test that formatting signature works on a static method.
        """
        myclass = treedoctestpackage.MyClass()

        assert (
            "".join(
                char
                for char in format_signature(
                    myclass.static_method_bound_to_myclass, verbosity=verbosity
                )
                if char != " "
            )
            == expected
        )

    @staticmethod
    def test_width_restriction():
        """Test that the width is always respected."""

        func = treedoctestpackage.module.func_many_long_args

        generator = itertools.product(list(range(15, 100)), [0, 1, 2, 3, 4])
        for width, verbosity in generator:
            formatted_signature = format_signature(
                func, width=width, verbosity=verbosity
            )
            assert len(formatted_signature) <= width


if __name__ == "__main__":
    import pytest

    pytest.main(args=[__file__, "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Utility functions which are not in use.
"""

import inspect


def is_method(obj):
    """Whether an object is a method or not."""
    return inspect.ismethoddescriptor(obj) or inspect.ismethod(obj)


def is_bound_method(obj):
    """Whether a method is bound to a class or not."""
    condition1 = "." in obj.__qualname__
    if not inspect.getfullargspec(obj).args:
        return False
    condition2 = inspect.getfullargspec(obj).args[0] == "self"
    return condition1 and condition2


def inspect_classify(obj):
    """
    Classify an object according to the inspect module. Not disjoint.
    
    Examples
    --------
    >>> inspect_classify(list)
    ['class']
    >>> f = lambda x : x * x
    >>> inspect_classify(f)
    ['function', 'routine']
    >>> class Vector:
    ...     def add(self, other):
    ...         pass
    >>> inspect_classify(Vector)
    ['class']
    >>> inspect_classify(Vector.add)
    ['function', 'routine']
    >>> gen = (i for i in range(10))
    >>> inspect_classify(gen)
    ['generator']
    >>> inspect_classify(max)
    ['builtin', 'routine']
    
    
    Common findings:
        []
        ['abstract', 'class']
        ['builtin', 'routine']
        ['class']
        ['datadescriptor']
        ['datadescriptor', 'getsetdescriptor']
        ['datadescriptor', 'memberdescriptor']
        ['function', 'generatorfunction', 'routine']
        ['function', 'routine']
        ['methoddescriptor', 'routine']
        ['method', 'routine']
        ['module']
        
        Numpy ufuncs not found.
    """
    classes = list()

    for function_name in sorted(dir(inspect)):

        if not function_name.startswith("is"):
            continue

        function = getattr(inspect, function_name)

        if function(obj):
            classes.append(function_name[2:])

    return classes


if __name__ == "__main__":
    import pytest

    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=[__file__, "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 21 21:13:23 2019

@author: tommy
"""
import functools


def function_nested_outer(x):
    def function_nested_inner(x):
        return x

    return function_nested_inner


def function_with_inner_class(x):
    class ClassInsideFunction:
        pass

    return x


def function_which_will_become_staticmethod():
    pass


def wrapper(function):
    @functools.wraps(function)
    def wrapped(*args, **kwargs):
        return function(*args, **kwargs)

    return wrapped


@wrapper
def func_wrapped(a, b=2):
    return a + b


# =============================================================================
# CLASSES AND INHERITANCE
# =============================================================================


class SuperClass:
    def superclass_method(self):
        return 1

    static_method = staticmethod(function_which_will_become_staticmethod)


class SubClass(SuperClass):
    def subclass_method(self):
        return 1


class SubSubClass(SubClass):
    def subsubclass_method(self):
        return 1


# =============================================================================
# CLASSES WITH COMPOSITION
# =============================================================================


class Wheel:
    pass


class Car:

    wheel_cls = Wheel
    wheel = Wheel()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for printers (classes) and associated functions.
"""

import abc
import collections
import importlib
import inspect
import pkgutil
import pydoc
import sys
import textwrap
import typing

from treedoc.utils import Peekable, PrintMixin

# =============================================================================
# ------------------------ PART 1/2 OF MODULE - CLASSES -----------------------
# =============================================================================


class PrinterABC(abc.ABC):
    """Abstract base class (ABC) for printers."""

    @abc.abstractmethod
    def __init__(self):
        pass

    @abc.abstractmethod
    def format_iterable(self, iterable) -> typing.Generator[str, None, None]:
        """Consumes an iterator yielding object stacks and yields formatted strings.
        
        Parameters
        ----------
        iterator : Iterator
            An iterator yielding (print_stack, stack). 
            The stacks are lists representing paths in the object tree.
        
        Yields
        ------
        formatted_str : string
            A string to be printed
        """
        pass


class Printer(PrintMixin):
    """Base class for printers, used for input validation."""

    def __init__(
        self, *, signature: int = 1, docstring: int = 2, info: int = 2, width: int = 88
    ):
        """Initialize a printer.
        
        The interpretation of the following arguments varies from printer to printer.
        
        Parameters
        ----------
            signature : int
                How much signature information to show
            docstring : int
                How much docstring information to show
            info : int
                How much general information to show
            width : int
                Maximum character width

        """
        assert signature in (0, 1, 2, 3, 4)
        assert docstring in (0, 1, 2)
        assert info in (0, 1, 2, 3, 4)
        assert width in range(2 ** 6, 2 ** 8 + 1)
        self.signature = signature
        self.docstring = docstring
        self.info = info
        self.width = width

    def _validate_obj_stack(self, stack) -> None:
        assert isinstance(stack, collections.abc.Sequence)
        assert len(stack) > 0
        assert all((hasattr(obj, "__name__") for obj in stack))


class DensePrinter(Printer, PrinterABC):

    SEP = " -> "
    END = "\n"
    SPACES2 = "  "
    SPACES4 = 2 * SPACES2

    def _get_docstring(self, obj) -> str:
        """Get and format the docstring of an object."""
        if self.docstring == 0:
            return ""
        return get_docstring(obj, width=self.width)

    def _format_signature(self, obj) -> str:
        """Get and format the signature of an object."""
        formatted = format_signature(obj, verbosity=self.signature, width=self.width)
        assert isinstance(formatted, (str,))
        return formatted

    def format_iterable(self, iterable) -> typing.Generator[str, None, None]:
        """ See the Abstract Base Class for the docstring.
        
        Summary
        -------
        Take an iterable object yielding (stack, final_node_at_depth) and returns strings.
        """
        for stack in self._format_row(iterable):

            self._validate_obj_stack(stack)

            # The row represents a path in the tree, the leaf object is the "final" object
            *_, last_obj = stack

            # Attempt to get a signature for the last object
            signature = self._format_signature(last_obj)

            # Get docstring information from the leaf object
            docstring = self._get_docstring(last_obj)

            # Info determines how much to show
            if self.info == 0:
                obj_names = stack[-1].__name__
            else:
                obj_names = ".".join([s.__name__ for s in clean_object_stack(stack)])

            yield (
                "{}|".format(describe(last_obj))
                + self.SPACES2 * len(stack)
                + obj_names
                + signature
                + self.SPACES4
                + docstring
            )

    def _format_row(self, iterable):
        """Yield the object stack from the iterable."""
        for stack, final_node_at_depth in iterable:
            yield stack


class TreePrinter(Printer, PrinterABC):

    # Class attributes used for printing
    RIGHT = "├──"
    DOWN = "│  "
    LAST = "└──"
    BLANK = "   "

    def _get_docstring(self, obj, *, width) -> str:
        """Get and format the docstring of an object."""
        if self.docstring == 0:
            return ""

        return get_docstring(obj, width=width)

    def _format_signature(self, obj, *, width) -> str:
        """Get and format the signature of an object."""
        formatted = format_signature(obj, verbosity=self.signature, width=width)
        assert isinstance(formatted, str)
        return formatted

    def format_iterable(self, iterable) -> typing.Generator[str, None, None]:
        """ See the Abstract Base Class for the docstring.
        
        Summary
        -------
        Take an iterable object yielding (stack, final_node_at_depth) and returns strings.
        """
        assert isinstance(iterable, collections.abc.Iterable)

        for print_stack, stack in self._format_row(iterable):

            self._validate_obj_stack(stack)

            joined_print_stack = " ".join(print_stack)

            # Info determines how much to show
            if self.info == 0:
                obj_names = stack[-1].__name__
            else:
                obj_names = ".".join([s.__name__ for s in clean_object_stack(stack)])

            # TODO: Consider a check of whether the object can be resolved here

            # TODO: Differentiate between INFO = 1 AND INFO = 2

            # The 1 represents the spacing between the stacks
            # The 3 represents space and quotation marks in "docstring"
            width_used = len(joined_print_stack) + len(obj_names) + 1

            last_obj = stack[-1]
            signature = self._format_signature(last_obj, width=self.width - width_used)
            docstring = self._get_docstring(
                last_obj, width=self.width - len(joined_print_stack) - 3
            )

            to_yield = " ".join([joined_print_stack, obj_names]) + signature
            assert len(to_yield) <= self.width

            yield to_yield

            # No need to show docstring, or no docstring to show, simply continue
            if self.docstring == 0 or docstring == "":
                continue

            # Want to print with docstring on the new line. Logic to switch up symbols
            last_in_stack = print_stack[-1]

            if last_in_stack == self.RIGHT:
                symbol = self.DOWN
            elif last_in_stack == self.LAST:
                symbol = self.BLANK
            else:
                symbol = ""

            print_stack[-1] = symbol
            to_yield = " ".join([" ".join(print_stack), '"{}"'.format(docstring)])
            assert len(to_yield) <= self.width

            yield to_yield

    def _format_row(self, iterator, depth=0, print_stack=None):
        """Takes an iterator and yields tuples (print_stack, stack).
        
        This recursive generator takes an iterator which yields 
        (stack, final_node_at_depth) and from it computes (print_stack, stack).
        It's purpose is to generate the pretty tree-structure from the 
        `final_node_at_depth` stack.
        """

        if not isinstance(iterator, Peekable):
            iterator = Peekable(iter(iterator))

        print_stack = print_stack or [""]

        # =============================================================================
        #         BOUNDARY CONDITIONS - For the root node and leaf nodes
        # =============================================================================

        # Peek to see if this is the final note. If it is, return
        stack, final_node_at_depth = iterator.peek((False, False))
        if not stack and not final_node_at_depth:
            return

        # Special case for the root note
        if len(stack) == 1:
            yield print_stack, stack
            next(iterator)
            yield from self._format_row(
                iterator, depth=depth + 1, print_stack=print_stack
            )
            return

        # =============================================================================
        #         ITERATE OVER NODES - Iteration and recursion if needed
        # =============================================================================

        # Iterate over every child at this level
        while True:

            # Get new elements from the iterator, default is (False, False)
            stack, final_node_at_depth = next(iterator, (False, False))

            # We reached the end of the tree, and must go up in the structure
            if not stack and not final_node_at_depth:
                return

            # At the current depth, is this node the final one?
            final_at_depth = final_node_at_depth[depth]

            # Yield the current node at the current depth
            symbol = self.LAST if final_at_depth else self.RIGHT
            yield print_stack + [symbol], stack

            # TODO: IMPORTANT: Calling `peek` will override the `stack` and
            # `final_node_at_depth` variables. I am not sure why.
            # But we need to get all information from these before the `peek` call.
            len_stack = len(stack)

            next_stack, next_final_node_at_depth = iterator.peek((False, False))

            # We reached the end of the tree, and must go up in the structure
            if not next_stack and not next_final_node_at_depth:
                return

            # The next stack has more elements, so it's a child of the current node
            if len(next_stack) > len_stack:
                # Find the appropriate recursion symbol and then recurse
                rec_symbol = self.BLANK if final_at_depth else self.DOWN

                yield from self._format_row(
                    iterator,
                    depth=depth + 1,
                    print_stack=print_stack.copy() + [rec_symbol],
                )

                # The node was the final one at this level, so we go up in the tree
                if final_at_depth:
                    return
            elif len(next_stack) < len_stack:
                # The node was the final one at this level, so we go up in the tree
                return
            else:
                # The next stack is of the same length as the current one
                continue


# =============================================================================
# ------------------------ PART 2/2 OF MODULE - FUNCTIONS ---------------------
# =============================================================================


def _describe(obj) -> str:
    """Produce a short description of the given object.
    
    Inspired by pydoc.describe.
    """
    if inspect.ismodule(obj):

        if obj.__name__ in sys.builtin_module_names:
            return "built-in module"

        if hasattr(obj, "__path__"):
            return "package"
        else:
            return "module"

    if inspect.isbuiltin(obj):
        return "built-in function"

    if inspect.isgetsetdescriptor(obj):
        return "getset descriptor"

    if inspect.ismemberdescriptor(obj):
        return "member descriptor"

    if inspect.isclass(obj):
        return "class"

    if inspect.isfunction(obj):
        return "function"

    if inspect.ismethod(obj):
        return "method"

    return type(obj).__name__


def describe(obj) -> str:
    """Produce a short description of the given object."""
    return _describe(obj).ljust(17)


def get_docstring(obj, *, width=88) -> str:
    """Get a docstring summary from an object.
    
    If no docstring is available, an empty string is returned.
    
    Examples
    --------
    >>> get_docstring(set.intersection)
    'Return the intersection of two sets as a new set.'
    >>> get_docstring(set.intersection, width=18)
    'Return the...'
    """

    # =============================================================================
    #     TODO: This function needs to become smarter, i.e. more adaptible to
    #     how docstrings are written "in the wild".
    # =============================================================================

    # pydoc.getdoc is slightly more general than inspect.getdoc,see:
    # https://github.com/python/cpython/blob/master/Lib/pydoc.py#L92
    doc = pydoc.getdoc(obj)
    first_line, _ = pydoc.splitdoc(doc)

    return_str = textwrap.shorten(first_line, width=width, placeholder="...")

    if return_str:
        return_str = return_str[:-3] if return_str.endswith("....") else return_str
        return return_str

    # If the docstring is a long paragraph, pydoc.splitdoc will return ''
    # We change this behavior to include the start of the string.
    if not return_str and doc:
        return_str = textwrap.shorten(doc, width=width, placeholder="...")
        return_str = return_str[:-3] if return_str.endswith("....") else return_str
        return return_str

    return ""


def clean_object_stack(stack):
    """
    Join an object stack so that consecutive modules are merged into the last one.
    
    This avoids the stack [collections, collections.abc] as being named
    'collections.collections.abc'.
    
    >>> import collections
    >>> from collections import abc
    >>> clean_object_stack([collections, abc]) == [abc]
    True
    >>> from collections import Counter
    >>> input_stack = [collections, Counter, Counter.most_common]
    >>> clean_object_stack(input_stack) == input_stack
    True
    """
    assert isinstance(stack, list)

    stack = stack.copy()
    new_stack = []

    for obj in stack:
        # This one is a module, and the last one is a module
        # Therefore it must be a submodule, get rid of the first one
        # to avoid [collections, collections.abc]
        if new_stack and inspect.ismodule(obj) and inspect.ismodule(new_stack[-1]):
            new_stack.pop()
        new_stack.append(obj)

    assert len(new_stack) > 0
    return new_stack


def _get_name(param) -> str:
    """Checks if signature.Parameter corresponds to *args or **kwargs type input."""
    if (
        param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD)
        and param.default is param.empty
    ):
        return str(param)
    else:
        return param.name


def _between(string, start, end) -> str:
    """Returns what's between `start` and `end`, exclusive.
    
    Examples
    >>> _between("im a nice (person) to all", "(", ")")
    'person'
    >>> _between("im a nice (person to all", "(", ")")
    Traceback (most recent call last):
        ...
    ValueError: substring not found
    """
    i = string.index(start)
    part = string[i + len(start) :]
    j = part.index(")")
    return part[:j]


def signature_from_docstring(obj):
    """Extract signature from built-in object docstring.
    
    Some of the built-in methods have signature info in the docstring. One example is
    `dict.pop`. This method will retrieve the docstring, and return None if it cannot.
    
    >>> import math
    >>> signature_from_docstring(math.log) in ('x[, base]', 'x, [base=math.e]')
    True
    >>> signature_from_docstring(dict.pop) in ('k[,d]', None)
    True
    """

    # If not docstring is available, return
    docstring_line = get_docstring(obj)

    if not docstring_line:
        return None

    # If it's not callable, return
    if not isinstance(obj, collections.abc.Callable):
        return None

    # Look for the name of the object, i.e. func(x)
    assert hasattr(obj, "__name__")

    if not obj.__name__ in docstring_line:
        return None

    # Attempt to get the signature
    index = docstring_line.index(obj.__name__) + len(obj.__name__)
    signature_part = docstring_line[index:]

    try:
        ret_value = _between(signature_part, "(", ")")
        if ret_value:
            return ret_value
    except ValueError:
        return None

    return None


def format_signature(obj, *, verbosity=2, width=88) -> str:
    """ 
    Format a function signature for printing.
    
    This function tries first to use inspect.signature, if that fails it will look for
    signature information in the first line of the docstring, and if that fails it will
    return a generic sigature if the object is callable.
    
    Examples
    --------
    >>> import collections
    >>> # Works on functions with well-defined signature
    >>> format_signature(collections.defaultdict.fromkeys, verbosity=2)
    '(iterable, value)'
    >>> # Built-ins with signature information in the docs
    >>> format_signature(collections.defaultdict.update, verbosity=2)
    '([E, ]**F)'
    >>> # Built-ins with signature no signature information at all
    >>> format_signature(collections.deque.append, verbosity=2)
    '(...)'
    """
    signature_string = _format_signature(obj=obj, verbosity=verbosity)
    assert isinstance(signature_string, str)

    # No problems with the width
    if len(signature_string) <= width:
        return signature_string

    # It's too wide, shorten it and return
    inner_sig = str(signature_string).strip("()")

    try:
        inner_sig = textwrap.shorten(inner_sig, width=width - 2, placeholder=" ...")
    except ValueError:  # ValueError: placeholder too large for max width
        inner_sig = ""

    result = "(" + inner_sig + ")"

    assert len(result) <= width

    return result


def _format_signature(obj, *, verbosity=2) -> str:
    """Format the signature, but make no guarantee for width."""
    # TODO: Figure out how to handle *

    max_verbosity = 4
    assert 0 <= verbosity <= max_verbosity
    SEP = ", "

    # Return formatted signature based on verbosity
    if verbosity == 0:
        return ""

    # Check if object has signature
    try:
        sig = inspect.signature(obj)
    except ValueError:
        # inspect.signature raises ValueError if no signature can be provided.
        # Example:
        # -------
        # >>> inspect.signature(math.log)
        # 'ValueError: no signature found for builtin <built-in function log>'

        signature_in_docs = signature_from_docstring(obj)

        # Failed to find a signature in the docstring
        if signature_in_docs is None:
            # inspect.signature and docstring info has failed, still return if callable
            if callable(obj) and verbosity >= 1:
                return "(...)"
            return ""
        # Found a signature in the docstring
        else:
            if verbosity == 1:
                return "(...)"
            else:
                # TODO: Format this more depending on verbosity
                return "(" + signature_in_docs + ")"
        return ""

    except TypeError:
        # inspect.signature raises TypeError if type of object is not supported.
        # Example:
        # -------
        # >>> x = 1
        # >>> inspect.signature(x)
        # 'TypeError: 1 is not a callable object'
        return ""

    # =============================================================================
    #     If the code reaches this point, we have a Signature stored in `sig`
    # =============================================================================

    # If function as no arguments, return
    if str(sig) == "()" and verbosity > 0:
        return str(sig)

    # Check if signature has annotations or defaults
    annotated = any(
        param.annotation is not param.empty for param in sig.parameters.values()
    )
    has_defaults = any(
        param.default is not param.empty for param in sig.parameters.values()
    )

    # Dial down verbosity if user has provided a more verbose alternative than is available
    if not annotated:
        max_verbosity = 3

    if not has_defaults and not annotated:
        max_verbosity = 2

    if verbosity > max_verbosity:
        # TODO: Give user warning if verbosity needs to be adjusted, e.g.
        # print(f'Adjusting verbosity: {verbosity} -> {max_verbosity}.')
        verbosity = max_verbosity
        # Fails for instance on collections.ChainMap without this logic
        return str(sig)

    elif verbosity == 1:
        return "(...)"

    elif verbosity == 2:
        return_sig = SEP.join(_get_name(param) for param in sig.parameters.values())
        return "(" + return_sig + ")"

    elif verbosity == 3:
        return_sig = SEP.join(
            param.name + "=" + str(param.default)
            if param.default is not param.empty
            else _get_name(param)
            for param in sig.parameters.values()
        )
        return "(" + return_sig + ")"

    else:
        return str(sig)


def resolve_str_to_obj(object_string) -> object:
    """
    Resolve a string to a Python object.
    
    list -> builtin
    collections -> module
    collections.Counter -> object
    
    Examples
    --------
    >>> resolve_str_to_obj("list") == list
    True
    >>> import collections
    >>> resolve_str_to_obj("collections") == collections
    True
    >>> from collections import Counter
    >>> resolve_str_to_obj("collections.Counter") == Counter
    True
    >>> resolve_str_to_obj("gibberish.Counter")
    Traceback (most recent call last):
        ...
    ImportError: Could not resolve 'gibberish.Counter'.
    """
    assert isinstance(object_string, str)
    suggestion = pydoc.locate(object_string)

    if suggestion is None and object_string != "None":
        raise ImportError("Could not resolve '{}'.".format(object_string))
    else:
        return suggestion


def resolve_input(obj):
    """Resolve a general input (str, iterable, etc) to a list of Python objects.
    
    Examples
    --------
    >>> resolve_input("list") == [list]
    True
    >>> resolve_input("list dict") == [list, dict]
    True
    >>> len(resolve_input(["python"])) > 10
    True
    >>> resolve_input(["list"]) == [list]
    True
    >>> resolve_input(["list", "dict"]) == [list, dict]
    True
    >>> resolve_input(list) == [list]
    True
    >>> from collections.abc import Collection
    >>> resolve_input("collections.abc.Collection") == [Collection]
    True
    """
    if isinstance(obj, str):
        obj = obj.strip()

    if isinstance(obj, str) and " " in obj:
        # Parse "dict   str  " correctly, ignoring extra whitespace
        obj = [o for o in obj.split(" ") if o != ""]

    # Special handling if "python" is passed
    if isinstance(obj, list) and len(obj) == 1 and obj[0].lower().strip() == "python":
        objects = []
        for (importer, object_name, ispkg) in pkgutil.iter_modules():
            if not ispkg:
                continue
            try:
                obj = importlib.import_module(object_name)
            except:
                continue
            objects.append(obj)
        return objects
    elif isinstance(obj, str):
        assert " " not in obj
        return [resolve_str_to_obj(obj)]
    elif isinstance(obj, (list, set, tuple)):
        attempt = [resolve_str_to_obj(o) if isinstance(o, str) else o for o in obj]
        return [obj for obj in attempt if obj is not None]
    else:
        return [obj]

    raise ValueError("Could not resolve object")


if __name__ == "__main__":
    import pytest

    pytest.main(args=[".", "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Tests for classes and functions located in `traversal.py`.
"""

import inspect
import itertools
import operator

import pytest

import treedoctestpackage as testpackage
import treedoctestpackage.subpackage as subtestpackage
from treedoc.traversal import ObjectTraverser, descend_from_package, is_package
from treedoctestpackage import module


class TestObjectTraverser:
    @staticmethod
    @pytest.mark.parametrize("modules", [True, False])
    def test_recusion_subpackages(modules):
        """Regardless of what the `modules` flag says, the following should hold."""

        traverser = ObjectTraverser(subpackages=True, modules=modules)
        assert traverser.recurse_to_child_object(
            obj=testpackage, child_obj=subtestpackage
        )

        traverser = ObjectTraverser(subpackages=False, modules=modules)
        assert not traverser.recurse_to_child_object(
            obj=testpackage, child_obj=subtestpackage
        )

    @staticmethod
    @pytest.mark.parametrize(
        "subpackages, modules", itertools.product([True, False], [True, False])
    )
    def test_recusion_subpackages_never_up(subpackages, modules):
        """As a fail-safe, we assure that recursion will never go up."""

        traverser = ObjectTraverser(subpackages=subpackages, modules=modules)
        # Notice that the arguments are switched and in the wrong order
        assert not traverser.recurse_to_child_object(
            obj=subtestpackage, child_obj=testpackage
        )

    @staticmethod
    @pytest.mark.parametrize(
        "subpackages, modules", itertools.product([True, False], [True, False])
    )
    def test_recusion_never_other_packages(subpackages, modules):
        """While inspect.getmembers(module) yields `operator` if `import operator` is
        present in `module.py`, we should never recurse to that child."""

        traverser = ObjectTraverser(subpackages=subpackages, modules=modules)
        # Notice that the arguments are switched and in the wrong order
        assert not traverser.recurse_to_child_object(obj=module, child_obj=operator)

    @staticmethod
    @pytest.mark.parametrize("subpackages", [True, False])
    def test_recusion_modules(subpackages):
        """Regardless of what the `subpackages` flag says, the following should hold."""

        traverser = ObjectTraverser(subpackages=subpackages, modules=True)
        assert traverser.recurse_to_child_object(obj=testpackage, child_obj=module)

        traverser = ObjectTraverser(subpackages=subpackages, modules=False)
        assert not traverser.recurse_to_child_object(obj=testpackage, child_obj=module)

    @staticmethod
    @pytest.mark.parametrize(
        "subpackages, modules", itertools.product([True, False], [True, False])
    )
    def test_recusion_to_objs_only_where_defined(subpackages, modules):
        """Do not recurse to objects defined elsewhere (unless we're in __init__.py).
        """
        from treedoctestpackage import module
        from treedoctestpackage.module2 import SuperClass

        traverser = ObjectTraverser(subpackages=subpackages, modules=modules)
        assert not traverser.recurse_to_child_object(obj=module, child_obj=SuperClass)

    @staticmethod
    @pytest.mark.parametrize("subpackages", [True, False])
    def test_recursion_objs_same_level(subpackages):
        """Test discover of object imported into __init__.py, defined at same level.
        
        Consider the structure:
            __init__.py (`MyClass` imported into here)
            module.py (`MyClass` defined here)
            
        If recursion to modules is activated, we discover module.py:MyClass.
        If recursion to modules it NOT activated, we discover it as __init__.py:MyClass.
        """
        import treedoctestpackage
        from treedoctestpackage.module import MyClass

        # =============================================================================
        #     `MyClass` is defined at the same level, not in __init__.py, but imported
        # =============================================================================

        # Recurse if it would not have been found elsewhere and it's in __init__.py
        assert MyClass in [obj for (_, obj) in inspect.getmembers(treedoctestpackage)]

        traverser = ObjectTraverser(subpackages=subpackages, modules=False)
        assert traverser.recurse_to_child_object(
            obj=treedoctestpackage, child_obj=MyClass
        )

        # It would've been found eventually, so don't discover it in __init__.py
        traverser = ObjectTraverser(subpackages=subpackages, modules=True)
        assert not traverser.recurse_to_child_object(
            obj=treedoctestpackage, child_obj=MyClass
        )

    @staticmethod
    @pytest.mark.parametrize("modules", [True, False])
    def test_recursion_objs_lower_level(modules):
        """Test discover of object imported into __init__.py, defined at at lower level.
        
        Consider the structure:
            __init__.py (`func_subtraction` imported into here)
            subpackage/subpackagemodule.py (`func_subtraction` defined here)
            
        If recursion to subpackages is activated, we discover 
            subpackage/subpackagemodule.py:func_subtraction.
            
        If recursion to subpackages it NOT activated, we discover 
            it as __init__.py:func_subtraction.
        """
        import treedoctestpackage
        from treedoctestpackage.subpackage.subpackagemodule import func_subtraction

        # =============================================================================
        #     `func_subtraction` is defined at a lower level, and import in __init__.py
        # =============================================================================

        # Recurse if it would not have been found elsewhere and it's in __init__.py
        assert func_subtraction in [
            obj for (_, obj) in inspect.getmembers(treedoctestpackage)
        ]

        traverser = ObjectTraverser(subpackages=False, modules=modules)
        assert traverser.recurse_to_child_object(
            obj=treedoctestpackage, child_obj=func_subtraction
        )

        # It would've been found eventually, so don't discover it in __init__.py
        traverser = ObjectTraverser(subpackages=True, modules=modules)
        assert not traverser.recurse_to_child_object(
            obj=treedoctestpackage, child_obj=func_subtraction
        )

    @staticmethod
    @pytest.mark.parametrize(
        "subpackages, modules", itertools.product([True, False], [True, False])
    )
    def test_recursion_composite_classes(subpackages, modules):
        """We do not recurse to composite classes, i.e. from Car to it's Wheel.
        Instead we find the wheel class from it's module."""
        from treedoctestpackage.module2 import Car, Wheel
        from treedoctestpackage import module2

        traverser = ObjectTraverser(subpackages=subpackages, modules=modules)
        assert traverser.recurse_to_child_object(obj=module2, child_obj=Car)
        assert traverser.recurse_to_child_object(obj=module2, child_obj=Wheel)

        assert not traverser.recurse_to_child_object(obj=Car, child_obj=Wheel)


def map_itemgetter(iterable, index: int):
    """Map an itemgetter over an iterable, returning element correpoding to index."""
    getter = operator.itemgetter(index)
    for item in iterable:
        yield getter(item)


def test_is_package():

    import treedoctestpackage

    assert is_package(treedoctestpackage)

    from treedoctestpackage import subpackage

    assert is_package(subpackage)

    from treedoctestpackage import module

    assert not is_package(module)

    from treedoctestpackage.subpackage import subpackagemodule

    assert not is_package(subpackagemodule)


class TestDescendFromPackage:
    @staticmethod
    def test_package_to_subpackages():
        """
        Test that yielding subpackages one level down works.
        """

        import treedoctestpackage

        sub_packages = descend_from_package(
            treedoctestpackage, include_subpackages=True
        )
        sub_packages = set(map_itemgetter(sub_packages, 1))

        from treedoctestpackage import subpackage, subpackage2

        assert set([subpackage, subpackage2]) == sub_packages

        # We only go one subpackage down
        from treedoctestpackage.subpackage import subsubpackage

        assert subsubpackage not in sub_packages

        # If we go down from a subpackage, we get a subsubpackage

        from treedoctestpackage import subpackage

        sub_packages = descend_from_package(subpackage, include_subpackages=True)
        sub_packages = set(map_itemgetter(sub_packages, 1))
        assert set([subsubpackage]) == sub_packages

    @staticmethod
    def test_package_to_modules():
        """
        Test that yielding modules one level down works.
        """
        import treedoctestpackage

        modules = descend_from_package(treedoctestpackage, include_modules=True)
        modules = set(map_itemgetter(modules, 1))

        from treedoctestpackage import module, module2

        assert set([module, module2]) == modules

    @staticmethod
    def test_package_to_modules_w_private():
        """
        Test that yielding modules one level down works with hidden modules too.
        """
        import treedoctestpackage

        modules = descend_from_package(
            treedoctestpackage, include_private=True, include_modules=True
        )
        modules = set(map_itemgetter(modules, 1))

        from treedoctestpackage import module, module2, _hidden_module

        assert set([module, module2, _hidden_module]) == modules


if __name__ == "__main__":
    import pytest

    pytest.main(args=[".", "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 21 21:13:23 2019

@author: tommy
"""

import operator


def func_addition(a, b):
    """Permforms addition."""
    return operator.add(a, b)


def func_subtraction(a, b):
    """Permforms addition."""
    return operator.subtract(a, b)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Tests for classes and functions located in `utils.py`.
"""


if __name__ == "__main__":
    import pytest

    # --durations=10  <- May be used to show potentially slow tests
    pytest.main(args=[__file__, "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for the ObjectTraverser class and associated functions.


TODO: Description.
                   +---+                  
      -------------|obj|--------------    
      |            +---+             |    
      |              |               |    
      v              v               v    
+----------+   +----------+   +----------+
|child_obj1|   |child_obj2|   |child_obj3|
+----------+   +----------+   +----------+


"""
import functools
import importlib
import inspect
import itertools
import os
import pkgutil
import sys

from treedoc.utils import PrintMixin

# Defining inspection funcs in global scope speeds up treedoc program by ~20%
_func_names = [func_name for func_name in dir(inspect) if func_name.startswith("is")]
INSPECTION_FUNCS = [getattr(inspect, func_name) for func_name in _func_names]
del _func_names

# =============================================================================
# ------------------------ PART 1/2 OF MODULE - CLASSES -----------------------
# =============================================================================


class ObjectTraverser(PrintMixin):
    """Traverse Python objects, modules and packages recursively."""

    _ignored_names = set(["__class__", "__doc__", "__hash__", "builtins", "__cached__"])

    def __init__(
        self,
        *,
        level=999,
        subpackages=False,
        modules=False,
        private=False,
        dunders=False,
        tests=False,
        stream=sys.stdout,
    ):
        self.level = level
        self.subpackages = subpackages
        self.modules = modules
        self.sort_key = None
        self.private = private
        self.dunders = dunders
        self.tests = tests
        self.stream = stream

    def search(self, obj):
        """DFS search starting at an object and recursing to its children."""
        yield from self._search(obj=obj, stack=None)

    def _p(self, *args):
        """Printing/logging method."""
        return None
        # TODO: set up proper logging
        print(*args, file=self.stream)

    def recurse_to_child_object(self, *, obj, child_obj) -> bool:
        """Given an object and its child, do we recurse down to the child?"""

        # TODO: Optimize this by placing what fails most often first

        self._p(f"obj = {obj.__name__}, child_obj = {child_obj.__name__}")

        # =============================================================================
        #  (1) CASE: Parent is a module, and child is not a module
        # =============================================================================

        if inspect.ismodule(obj) and not inspect.ismodule(child_obj):

            # Prevent `collections.eq` / `collections._eq`
            if inspect.isbuiltin(child_obj):
                if inspect.getmodule(child_obj) != obj:
                    self._p(f"OC: Failed on condition 1.1")
                    return False

            # Not defined in the sub-tree, skip it
            if not is_subpackage(inspect.getmodule(child_obj), obj):

                self._p(f"OC: Failed on condition 1.2")
                return False

            # The object is defined in a different file
            if inspect.getmodule(child_obj) != obj:

                # If the object is not __init__.py,
                # never include anything imported to it
                if not is_package(obj):
                    self._p(f"OC: Failed on condition 1.3")
                    return False

                # At this point the object is __init__.py, and we *might* include
                # a file imported into it (depending on settings below)

                # The object is defined at a lower level
                if is_propersubpackage(inspect.getmodule(child_obj), obj):

                    if self.subpackages:
                        # will find it later, so skip it now
                        self._p(f"OC: Failed on condition 1.4")
                        return False

                # If the object is defined at the same level
                if is_subpackage(inspect.getmodule(child_obj), obj) and is_subpackage(
                    obj, inspect.getmodule(child_obj)
                ):

                    if self.modules:
                        # will find it later, so skip it now
                        self._p(f"OC: Failed on condition 1.5")
                        return False

        # =============================================================================
        #  (2) CASE: Both parent and child are modules, i.e. __init__.py or module.py
        # =============================================================================

        if inspect.ismodule(obj) and inspect.ismodule(child_obj):

            # The order is wrong, i.e. `main` in `main.subpackage` implies going up
            different_packages = child_obj.__package__ != obj.__package__
            # pytest.collect has __package__ == None
            if child_obj.__package__ is None:
                child_package_wrong = False

            else:
                child_package_wrong = child_obj.__package__ in obj.__package__

            if different_packages and child_package_wrong:
                self._p(f"OC: Failed on condition 2.1")
                return False

            if child_obj.__package__ is not None:
                # Prevents for instance `pandas` to recurse into `numpy`
                if not obj.__package__ in child_obj.__package__:
                    self._p(f"OC: Failed on condition 2.2")
                    return False

            # Another fail safe to prevent `mysubpackage` to recurse into `subpack`
            if hasattr(obj, "__path__") and hasattr(child_obj, "__path__"):
                # TODO
                pass

            # Fail safe to prevent recursing from `pkg/file.py` to `pkg/__init__.py`
            if hasattr(obj, "__file__") and hasattr(child_obj, "__file__"):

                obj_pth, obj_py_file = os.path.split(inspect.getfile(obj))
                child_obj_pth, child_obj_py_file = os.path.split(
                    inspect.getfile(child_obj)
                )
                if not obj_pth in child_obj_pth:
                    self._p(f"OC: Failed on condition 2.3")
                    return False

                if inspect.getfile(obj) == inspect.getfile(child_obj):
                    self._p(f"OC: Failed on condition 2.4")
                    return False

                if child_obj_py_file == "__init__.py" and obj_py_file != "__init__.py":
                    self._p(f"OC: Failed on condition 2.5")
                    return False

            if is_propersubpackage(child_obj, obj) and not self.subpackages:
                self._p(f"OC: Failed on condition 2.6")
                return False

            try:
                file = inspect.getfile(child_obj)
                if (not file.endswith("__init__.py")) and not self.modules:
                    self._p(f"OC: Failed on condition 2.7")
                    return False
            except TypeError:
                # TypeError: <module 'sys' (built-in)> is a built-in module
                pass

            if obj.__package__ == child_obj.__package__ and not self.modules:
                self._p(f"OC: Failed on condition 2.8")
                return False

        # =============================================================================
        #  (3) CASE: The child is a class
        # =============================================================================

        # We're dealing with a class imported from another library, skip it
        # TODO: Extend this to other objects?
        if inspect.isclass(child_obj):

            if not is_subpackage(inspect.getmodule(child_obj), obj):
                self._p(f"OC: Failed on condition 3.1")
                return False

            if obj in inspect.getmro(child_obj):
                self._p(f"OC: Failed on condition 3.2")
                return False

            # We prefer going from modules to classes, not from classes to classes
            if inspect.isclass(obj):
                self._p(f"OC: Failed on condition 3.3")
                return False

        # =============================================================================
        #         if not is_subpackage(inspect.getmodule(child_obj), obj):
        #             self._p(f"OC: Failed on condition 4.1")
        #             return False
        # =============================================================================

        return True

    def recurse_to_object(self, obj) -> bool:
        """Given an object, should we recurse down to it?"""

        name = obj.__name__

        # The following order is based on line profiling in an attempt to speed
        # up recursion
        # Inline comments original condition -> renamed condition indicate
        # renaming condition following profiling

        # 1.5 -> 1.1
        if is_dunder_method(obj) and not self.dunders:
            self._p(f"O: Failed on condition 1.1")
            return False

        # 1.4 -> 1.2
        if is_private(obj) and not self.private:
            self._p(f"O: Failed on condition 1.2")
            return False

        # 1.2 -> 1.3
        if obj is type:
            self._p(f"O: Failed on condition 1.3")
            return False

        # 1.1 -> 1.4
        if name in self._ignored_names:
            self._p(f"O: Failed on condition 1.4")
            return False

        # 1.6 -> 1.5
        if not is_inspectable(obj):
            self._p(f"O: Failed on condition 1.5")
            return False

        # 1.3 -> 1.6
        if is_test(obj) and not self.tests:
            self._p(f"O: Failed on condition 1.6")
            return False

        return True

    def _search(self, *, obj, stack=None, final_node_at_depth=None):
        """
        Missing docstring.
        """

        # =============================================================================
        #         (1) BOUNDARY CONDITIONS
        #         These conditions are triggered when the DFS reaches a leaf node in
        #         the object tree, when the search has reached its desired depth or
        #         at the root note of the object tree.
        # =============================================================================

        stack = stack or []
        final_node_at_depth = final_node_at_depth or [True]

        self._p(
            f"yield_data({obj}, stack={stack}), final_node_at_depth={final_node_at_depth}"
        )

        if len(stack) > self.level + 1:
            self._p(f"Max level reached. Aborting.")
            return

        assert len(stack + [obj]) == len(final_node_at_depth)
        yield stack + [obj], final_node_at_depth

        # If it's not a module/package or class, we don't bother getting children
        if not (inspect.ismodule(obj) or inspect.isclass(obj)):
            return

        # =============================================================================
        #         (2) FILTER THE CHILD OBJECTS
        #         For efficient tree-like printing it's crucial to know whether
        #         a child node is the last child to be visited *before* the
        #         DFS search recurses. If a, b and c are children of A, then
        #         the algorithm must discover that c is the final child of A
        #         before it recurses on a, b or c.
        # =============================================================================

        # The objects we will recurse on
        generator1 = descend_from_package(
            package=obj,
            include_tests=self.tests,
            include_private=self.private,
            include_modules=self.modules,
            include_subpackages=self.subpackages,
        )
        generator2 = inspect.getmembers(obj)

        def unique_first(gen1, gen2):
            """Chain generators, but only one unique name."""
            # Using a list would be faster, but modules are not hashable, so O(n) lookup
            seen_objects = []
            seen_names = set()
            for name, obj in itertools.chain(generator1, generator2):

                # If it's not a module it's not a problem, since generator2 only yield modules
                if not inspect.ismodule(obj):
                    yield name, obj
                    continue

                try:
                    if obj not in seen_objects:
                        seen_objects.append(obj)
                        seen_names.add(name)
                        yield name, obj

                # The comparison operation has been overwritten. Fall back to naming
                # This happen for instance with pandas NaT (Not a Time)
                except TypeError:

                    if name not in seen_names:
                        seen_objects.append(obj)
                        seen_names.add(name)
                        yield name, obj

                except:
                    continue

        generator = unique_first(generator1, generator2)

        # Iterate through children
        filtered = []
        for name, child_obj in sorted(generator, key=self.sort_key):

            self._p(f"Looking at {name}, {type(child_obj)}")

            try:
                getattr(child_obj, "__name__")

            except AttributeError:
                try:
                    setattr(child_obj, "__name__", name)

                except AttributeError:
                    # This is for everything to work with properties, e.g. pandas.DataFrame.T
                    # TODO: Figure out how to deal with properties
                    continue

            assert hasattr(child_obj, "__name__")

            # Check if we should skip the object by virtue of its properties
            if not self.recurse_to_object(obj=child_obj):
                continue

            # Check if we should skip the child object by virtue of its relationship
            # to the parent object
            if not self.recurse_to_child_object(obj=obj, child_obj=child_obj):
                continue

            filtered.append((name, child_obj))

        # =============================================================================
        #         (3) RECURSE ON CHILD OBJECTS
        #         For every child object that we're interested in, recurse.
        # =============================================================================

        for num, (name, child_obj) in enumerate(filtered, 1):

            last = True if len(filtered) == num else False
            yield from self._search(
                obj=child_obj,
                stack=stack.copy() + [obj],
                final_node_at_depth=final_node_at_depth.copy() + [last],
            )


# =============================================================================
# ------------------------ PART 2/2 OF MODULE - FUNCTIONS ---------------------
# =============================================================================


def is_inspectable(obj) -> bool:
    """An object is inspectable if it returns True for any of the inspect.is.. functions."""
    is_inspectable_by_func = any((func(obj) for func in INSPECTION_FUNCS))
    return is_inspectable_by_func or isinstance(obj, functools.partial)


def is_propersubpackage(package_a, package_b) -> bool:
    """
    Is A a proper subpackage or submodule of B?
    """
    try:
        path_a, _ = os.path.split(inspect.getfile(package_a))
        path_b, _ = os.path.split(inspect.getfile(package_b))

    except TypeError:
        # Is a built-in module
        return False

    return (path_b in path_a) and not (path_b == path_a)


def is_subpackage(package_a, package_b) -> bool:
    """
    Is A a subpackage or submodule of B?
    """
    try:
        path_a, _ = os.path.split(inspect.getfile(package_a))
        path_b, _ = os.path.split(inspect.getfile(package_b))

    except TypeError:
        # Is a built-in module
        # For instance: is_subpackage(builtins, builtins) should return True
        if package_a == package_b:
            return True

        return False

    return path_b in path_a


def is_dunder_method(obj) -> bool:
    """Is the method a dunder (double underscore), i.e. __add__(self, other)?"""
    assert hasattr(obj, "__name__")

    obj_name = obj.__name__

    return obj_name.endswith("__") and obj_name.startswith("__")


def is_private(obj) -> bool:
    """Is the object private, i.e. _func(x)?"""
    assert hasattr(obj, "__name__")

    obj_name = obj.__name__
    typical_private = obj_name.startswith("_") and obj_name[1] != "_"
    private_subpackage = "._" in obj_name

    return typical_private or private_subpackage


def is_test(obj) -> bool:
    """Is the object a test, i.e. test_func()?"""
    assert hasattr(obj, "__name__")

    obj_name = obj.__name__.lower()
    patterns = ("test", "_test", "__test")

    return any(obj_name.startswith(pattern) for pattern in patterns)


def is_package(obj) -> bool:
    """Does the object file end with '__init__.py'?"""
    if not hasattr(obj, "__file__"):
        return False

    return obj.__file__.endswith("__init__.py")


def descend_from_package(
    package,
    *,
    include_tests=False,
    include_private=False,
    include_modules=False,
    include_subpackages=False,
):
    """Descend one level down from a package to either a subpackage or modules.
    
    Yields a tuple of (object, object_name) one level down.
    """
    if not inspect.ismodule(package):
        return None

    try:
        path, _ = os.path.split(inspect.getfile(package))

    except TypeError:
        # Is a built-in module
        return None

    prefix = package.__name__ + "."

    generator = pkgutil.iter_modules(path=[path], prefix=prefix)

    for (importer, object_name, ispkg) in generator:

        ismodule = not ispkg

        # Covers names such as "test", "tests", "testing", ...
        if ".test" in object_name.lower() and not include_tests:
            continue

        if "._" in object_name.lower() and not include_private:
            continue

        try:
            obj = importlib.import_module(object_name)

        except ModuleNotFoundError:
            # TODO: Replace this with logging
            # print(f"Could not import {object_name}. Error: {error}")
            return

        except ImportError:
            # print(f"Could not import {object_name}. Error: {error}")
            return

        # File "/home/tommy/anaconda3/envs/treedoc/lib/python3.7/ctypes/wintypes.py", line 20, in <module>
        except ValueError:
            # print(f"Could not import {object_name}. Error: {error}")
            return

        # File "/home/tommy/anaconda3/envs/treedoc/lib/python3.7/ctypes/wintypes.py", line 20, in <module>
        except LookupError:
            # print(f"Could not import {object_name}. Error: {error}")
            return

        # File "/home/tommy/anaconda3/envs/treedoc/lib/python3.7/site-packages/numpy/ma/version.py", line 12, in <module>
        except AttributeError:
            # print(f"Could not import {object_name}. Error: {error}")
            return

        if include_subpackages and ispkg:
            yield object_name, obj

        if include_modules and ismodule:
            yield object_name, obj


if __name__ == "__main__":
    import pytest

    pytest.main(args=[".", "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
High level functions: treedoc-function, argument parsing and CLI entrypoint.
"""


import argparse
import sys

from treedoc.printing import DensePrinter, TreePrinter, resolve_input
from treedoc.traversal import ObjectTraverser
from treedoc.utils import get_terminal_size


def treedoc(
    obj,
    *,
    level=999,
    subpackages=False,
    modules=False,
    private=False,
    dunders=False,
    tests=False,
    signature=1,
    docstring=2,
    info=2,
    width=88,
    printer=TreePrinter,
    stream=sys.stdout,
):
    """Print minimalistic tree-like documentation.
    
    Arguments
    ---------
    obj
        A string representing an object, or several space-separated objects, or a list
        of objects. Examples: "dict", "dict set", [dict, set]
    
    Examples
    --------
    >>> treedoc(list)
    >>> treedoc("collections.Counter")
    """

    # A zero means that no width was explicitly set, if so we set it to terminal width
    if width == 0:
        width, _ = get_terminal_size(fallback=(128, 24))

    # Resolve the object
    objects = resolve_input(obj)

    # Pass the arguments to the object traverser and the printer
    traverser = ObjectTraverser(
        level=level,
        subpackages=subpackages,
        modules=modules,
        private=private,
        dunders=dunders,
        tests=tests,
        stream=stream,
    )
    printer = printer(signature=signature, docstring=docstring, info=info, width=width)

    for obj in objects:
        search_result_iterator = traverser.search(obj=obj)

        for row in printer.format_iterable(search_result_iterator):
            print(row, file=stream)


def setup_argumentparser(printers):
    """Set up the argument parser."""

    parser = argparse.ArgumentParser(
        prog="treedoc",  # The name of the program
        description="Minimalistic Python documentation for dendrophiles.",
        epilog="Report issues and contribute at https://github.com/tommyod/treedoc.",
        allow_abbrev=True,
        # add_help=True,
    )

    parser.add_argument(
        "obj",
        default="builtins",  # TODO: Change this to a different default?
        nargs="*",
        help="package/class/method/... , e.g. collections.Counter",
    )

    parser.add_argument(
        "-v",
        "--version",
        default=False,
        dest="version",
        action="store_true",
        help="print the version and exit.",
    )

    # =============================================================================
    #     OPTIONS RELATED TO OBJECT TRAVERSAL AND RECURSION
    # =============================================================================

    traversal = parser.add_argument_group(
        "traversal", "The arguments below are common to every printer."
    )
    traversal.add_argument(
        "-l",
        "--level",
        default=999,
        dest="level",
        nargs="?",
        type=int,
        help="descend only level directories deep.",
    )

    traversal.add_argument(
        "-s",
        "--subpackages",
        default=False,
        dest="subpackages",
        action="store_true",
        help="descend into subpackages, i.e. numpy -> numpy.linalg",
    )

    traversal.add_argument(
        "-m",
        "--modules",
        default=False,
        dest="modules",
        action="store_true",
        help="descend into every module in a package.",
    )

    traversal.add_argument(
        "-p",
        "--private",
        default=False,
        dest="private",
        action="store_true",
        help="show private objects, i.e. _func(x).",
    )

    traversal.add_argument(
        "-d",
        "--dunders",
        default=False,
        dest="dunders",
        action="store_true",
        help="show double underscore methods, i.e. __add__(self, other).",
    )

    traversal.add_argument(
        "-t",
        "--tests",
        default=False,
        dest="tests",
        action="store_true",
        help="show tests, i.e. test_func().",
    )

    # =============================================================================
    #     OPTIONS RELATED TO PRINTING THE RESULTS
    # =============================================================================

    printing = parser.add_argument_group(
        "printing",
        "The meaning of the arguments below may vary depending on the printer.\nNumeric arguments default to middle values, printer defaults to 'tree'.",
    )

    printing.add_argument(
        "-P",
        "--printer",
        default="tree",
        dest="printer",
        nargs="?",
        choices=list(printers.keys()),
        help="printer to use, defaults to 'tree'",
    )

    printing.add_argument(
        "-S",
        "--signature",
        action="store",
        default=2,
        dest="signature",
        type=int,
        choices=[0, 1, 2, 3, 4],
        help="how much signature information to show.",
    )

    printing.add_argument(
        "-D",
        "--docstring",
        action="store",
        default=1,
        dest="docstring",
        type=int,
        choices=[0, 1, 2],
        help="how much docstring information to show.",
    )

    printing.add_argument(
        "-I",
        "--info",
        action="store",
        default=2,
        dest="info",
        type=int,
        choices=[0, 1, 2, 3, 4],
        help="how much general information to show.",
    )

    printing.add_argument(
        "-W",
        "--width",
        action="store",
        default=0,
        dest="width",
        type=int,
        # choices=range(50, 500),
        help="maximal width of the output.",
    )

    return parser


def CLI_entrypoint():
    """Set up argumentparser and parse arguments, then pass arguments to treedoc."""
    printers = {"tree": TreePrinter, "dense": DensePrinter}
    parser = setup_argumentparser(printers)

    args = parser.parse_args()

    # The user wants to print the version. Print it and exit.
    if args.version:
        from treedoc import __version__

        print("treedoc version {}".format(__version__))
        return
    else:
        delattr(args, "version")

    args_to_func = {k: w for (k, w) in args._get_kwargs()}
    args_to_func["printer"] = printers[args_to_func["printer"]]

    treedoc(**args_to_func)


if __name__ == "__main__":
    import pytest

    pytest.main(args=[".", "--doctest-modules", "-v", "--capture=sys"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 21 21:13:23 2019

@author: tommy
"""

import functools
import operator


def func_addition(a, b):
    """Permforms addition."""
    return operator.add(a, b)


add_five_partial = functools.partial(func_addition, a=5)


def func_many_args(a, b=2, c=4, d=(1, 2, 3)):
    """Function with many arguments."""
    return b + c


def generator(a):
    for i in range(a):
        yield i


def wrapper(function):
    @functools.wraps(function)
    def wrapped(*args, **kwargs):
        return function(*args, **kwargs)

    return wrapped


@wrapper
def func_wrapped(a, b=2):
    return a + b


class MyClass:
    def __init__(self):
        pass

    def method_bound_to_myclass(self, a):
        """Method docstring."""
        return a

    @classmethod
    def classmethod_bound_to_myclass(cls, a):
        """Class method docstring."""
        return a

    @staticmethod
    def static_method_bound_to_myclass(self, a):
        """Static method docstring."""
        return a

    def __add__(self, other):
        return type(self)()



from os import path
import re
from setuptools import find_packages, setup

# Get the long description from README.md
PATH = path.abspath(path.dirname(__file__))
with open(path.join(PATH, "README.md"), encoding="utf-8") as file:
    long_description = file.read()

# Get version
with open(path.join(PATH, "treedoc/__init__.py"), encoding="utf-8") as file:
    version = re.search(r"__version__ = \"(.*?)\"", file.read()).group(1)

setup(
    name="treedoc",
    version=version,
    description="Minimalistic Python documentation for dendrophiles.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tommyod/treedoc",
    author="tommyod",
    author_email="tod001@uib.no",
    maintainer="tommyod, smu095, glemvik",
    maintainer_email="treedoc.dev@gmail.com",
    license="MIT",
    packages=find_packages(exclude=["*tests*"]),
    python_requires=">=3.6",
    install_requires=[],
    # https://python-packaging.readthedocs.io/en/latest/command-line-scripts.html
    entry_points={"console_scripts": ["treedoc = treedoc.main:CLI_entrypoint"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords=["documentation, utility"],
)
